/*------------------------------------------------------------------------
Module:        mime_alter.c
Author:        Paul L Daniels / Steven Coutts
Project:       ripMIME
State:         development
Creation Date: 01/05/2001
Description:   MIME-alter is an additional module in the mime.[ch] object, which will contain various functions which will -alter- mime-packs.
------------------------------------------------------------------------*/

// 0908 10H57 - Added ability for disclaimer to be actual text, as well as
// 	a filename.  I just detect for a \n or \r in the string, if detected
//		disclaimer is considered as "text"
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <syslog.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "ffget.h"
#include "pldstr.h"
#include "filename-filters.h"
#include "mime_alter.h"
#include "logger.h"

/* Our content types */
#define AM_contenttype_unknown -1
#define AM_contenttype_none 0
#define AM_contenttype_text_plain 1
#define AM_contenttype_text_html 2
#define AM_contenttype_text_print 3
#define AM_contenttype_text_multipart 4
#define AM_contenttype_text_unknown 5

/* Our encoding scemes */
#define AM_encoding_none 0
#define AM_encoding_base64 1
#define AM_encoding_print 2
#define AM_encoding_7bit 3


#define AM_1K_BUFFER_SIZE 1024
#define AM_DISCLAIMER_FILENAME_LENGTH_MAX 256;

#define AM_DNORMAL (glb.debug > 0)
#define AM_VERBOSE (glb.verbose > 0)

/* Globals */
char *AM_scratch_buffer[AM_1K_BUFFER_SIZE]; /* allocate a static array of 1024 pointers to char */


//06/03/2002 - replaced AM_setup_table() with this compile-time array. Saves a few startup cycles.

unsigned char AM_encode64[64]={
	65,   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   80,\
		81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   97,   98,   99,  100,  101,  102,\
		103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,\
		119,  120,  121,  122,   48,   49,   50,   51,   52,   53,   54,   55,   56,   57,   43,   47 \
};


#define AM_HEADERBUFFER_MAX 100
#define AM_HEADERBUFFER_ITEM_SIZE 1024

struct AM_globals {

	int debug;				// Low level debugging
	int verbose;				/* do we talk as we walk */
	int paranoid;				/* set paranoid to yes! */
	int HTML_too;				/* Add footer to the HTML email too */
	int multipart_insert;	/* Should we insert into emails which are embedded into another */
	int nullify_all;			/* Remove ALL filename'd attachments */
	int alter_signed;		/* Do we alter signed emails ? */

	char *disclaimer_plain;
	int disclaimer_plain_type;

	char *disclaimer_HTML;
	int disclaimer_HTML_type;

	char *headerbuffer[ AM_HEADERBUFFER_MAX ];	// 100 lines for the header buffers

	int headerbuffermax;
};

static struct AM_globals glb;




int AM_init( void )
{
	glb.debug=0;				// Low level debugging
	glb.verbose=0;				/* do we talk as we walk */
	glb.paranoid=1;				/* set paranoid to yes! */
	glb.HTML_too=1;				/* Add footer to the HTML email too */
	glb.multipart_insert=0;	/* do not insert into multipart attachments */
	glb.nullify_all=0;			/* Remove ALL filename'd attachments */
	glb.alter_signed=0;		/* Do we alter signed emails ? */

	glb.disclaimer_plain=NULL;
	glb.disclaimer_plain_type=AM_DISCLAIMER_TYPE_NONE;

	glb.disclaimer_HTML=NULL;
	glb.disclaimer_HTML_type=AM_DISCLAIMER_TYPE_NONE;

	glb.headerbuffermax=0;

	return 0;
}

/*------------------------------------------------------------------------
Procedure:     AM_set_debug ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_debug( int level )
{
	glb.debug = level;
	AM_set_verbose( level );

	return level;

}


/*------------------------------------------------------------------------
Procedure:     AM_hbuffer_reset ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_hbuffer_reset( void )
{
	int i;

	for (i = 0; i < glb.headerbuffermax; i++)
	{
		if (glb.headerbuffer[i]) free(glb.headerbuffer[i]);
	}
	glb.headerbuffermax=0;
	return 0;
}


/*------------------------------------------------------------------------
Procedure:     AM_hbuffer_add ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
char *AM_hbuffer_add( char *headerline, FILE *f )
{
	int i;

	if (glb.headerbuffermax == AM_HEADERBUFFER_MAX)
	{
		if (glb.headerbuffer[0])
		{
			fprintf(f,"%s",glb.headerbuffer[0]);
			free(glb.headerbuffer[0]);
		}
		for (i = 0; i < (AM_HEADERBUFFER_MAX-1); i++)
		{
			glb.headerbuffer[i] = glb.headerbuffer[i+1];
		}
		glb.headerbuffermax = AM_HEADERBUFFER_MAX -1;
	}

	glb.headerbuffer[glb.headerbuffermax] = malloc(sizeof(char) *(AM_HEADERBUFFER_ITEM_SIZE +1));

	if (!glb.headerbuffer[glb.headerbuffermax])
	{
		LOGGER_log("alterMIME: AM_hbuffer_add: Error: cannot allocate %d bytes for new header", AM_HEADERBUFFER_ITEM_SIZE +1);
		return NULL;
	}

	strncpy(glb.headerbuffer[glb.headerbuffermax], headerline, AM_HEADERBUFFER_ITEM_SIZE);
	glb.headerbuffermax++;

	return glb.headerbuffer[glb.headerbuffermax-1];
}


/*------------------------------------------------------------------------
Procedure:     AM_hbuffer_getmax ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_hbuffer_getmax( void )
{
	return glb.headerbuffermax;
}




/*------------------------------------------------------------------------
Procedure:     AM_hbuffer_getline ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
char *AM_hbuffer_getline( int index )
{
	if ((index >=0)&&(index < glb.headerbuffermax))
		return glb.headerbuffer[index];
	else return NULL;
}



/*------------------------------------------------------------------------
Procedure:     AM_set_glb.paranoid ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_paranoid( int level )
{
	glb.paranoid = level;
	return glb.paranoid;
}





/*------------------------------------------------------------------------
Procedure:     AM_set_glb.verbose ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_verbose( int level )
{
	glb.verbose = level;
	return glb.verbose;
}




/*------------------------------------------------------------------------
Procedure:     AM_set_HTMLtoo ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_HTMLtoo( int level )
{
	glb.HTML_too = level;
	return glb.HTML_too;
}


/*------------------------------------------------------------------------
Procedure:     AM_set_multipart_insert ID:1
Purpose:       Set the multipart-insertion of dislcaimers to level.  Multipart disclaimers are not on by default
because they would imply inserting disclaimers into previously forwarded messages.
Input:         level
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_multipart_insert( int level )
{
	glb.multipart_insert = level;
	return glb.multipart_insert;
}



char *AM_set_disclaimer_plain( char *filename, int disclaimer_type )
{
	glb.disclaimer_plain = strdup( filename );
	glb.disclaimer_plain_type = disclaimer_type;

	return glb.disclaimer_plain;
}



char *AM_set_disclaimer_HTML( char *filename, int disclaimer_type )
{
	glb.disclaimer_HTML = strdup( filename );
	glb.disclaimer_HTML_type = disclaimer_type;

	return glb.disclaimer_HTML;
}


/*------------------------------------------------------------------------
Procedure:     AM_set_nullifyall ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_nullifyall( int level )
{
	glb.nullify_all = level;
	return glb.nullify_all;
}



/*------------------------------------------------------------------------
Procedure:     AM_set_altersigned ID:1
Purpose:       Turns on or off the option to alter signed email messages.
Default is to be -off-.
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_altersigned( int level )
{
	glb.alter_signed = level;
	return glb.alter_signed;
}





/*------------------------------------------------------------------------
Procedure:     AM_base64_encodef ID:1
Purpose:       encode to BASE64 an input stream to an output stream
Input:         FILE *fin: Input stream
FILE *fout: Output stream
Output:
Errors:
------------------------------------------------------------------------*/
int AM_base64_encodef( FILE *fin, FILE *fout )
{
	unsigned char inbuf[3];
	unsigned char outbuf[4];
	int cc;
	int byte_count;

	if (!fin)
	{
		LOGGER_log("AM_base64_encodef: Error: input file stream not open, please use AM_base64_encode(infile,outfile)");
		return -1;
	}

	if (!fout)
	{
		LOGGER_log("AM_base64_encodef: Error: output file stream not open, please use AM_base64_encode(infile,outfile)");
		return -1;
	}

	cc = 0;	/* Character Count = 0 */

	while ((byte_count = fread(inbuf, 1, 3, fin)))	/* while we get more than 0 bytes */
	{
		/* Split 3 bytes into 4 bytes by taking 6bit blocks out of the 24 bit (3x8bit)
		 * array */
		outbuf[0] = AM_encode64[((inbuf[0] & 0xFC) >> 2)];
		outbuf[1] = AM_encode64[((inbuf[0] & 0x03) << 4) | ((inbuf[1] & 0xF0) >> 4)];
		outbuf[2] = AM_encode64[((inbuf[1] & 0x0F) << 2) | ((inbuf[2] & 0xC0) >> 6)];
		outbuf[3] = AM_encode64[(inbuf[2] & 0x3F)];

		/* if we didn't get a full 3 bytes from the file read, then we need to then
		 * pad the output with '=' (base64 padding char) */
		if ( byte_count < 3 )
		{

			if ( byte_count == 2 ) outbuf[3] = '=';
			if ( byte_count == 1 ) outbuf[3] = outbuf[2] = '=';

			fwrite(outbuf, 1, 4, fout); 	/* write out the buffer */
			fwrite( "\n", 1, 1, fout ); 	/* as this is the last line, put a trailing \n */
			break; 								/* exit out of the while loop */
		}

		fwrite(outbuf, 1, 4, fout);		/* normal operation buffer-write */

		cc += 4;									/* increment the output char count by 4 */

		if (cc > 76)							/* if we have more than 76 chars, then put in a \n */
		{
			fwrite( "\n", 1, 1, fout );	/* write the \n out */
			cc = 0;								/* reset the output char count */
		} /* if bytecount == 3 */
	} /* while more data to read in */

	return 0;
}






/*------------------------------------------------------------------------
Procedure:     AM_base64_encode ID:1
Purpose:       Encodes a given input stream into BASE64 and outputs to a given
output stream
Input:         char *enc_fname: Input filename of the stream to encode
char *out_fname: Output filename of the stream to send to
Output:
Errors:
------------------------------------------------------------------------*/
int AM_base64_encode( char *enc_fname, char *out_fname )
{
	FILE *fin, *fout;

	fin = fopen( enc_fname, "rb" );
	if (!fin)
	{
		LOGGER_log("AM_base64_encode: Cannot open \"%s\" for reading.(%s)",enc_fname,strerror(errno));
		return -1;
	}

	fout = fopen ( out_fname, "wb" );
	if (!fout)
	{
		LOGGER_log("AM_base64_encode: Cannot open \"%s\" for writing.(%s)",out_fname,strerror(errno));
		fclose(fin);
		return -1;
	}

	AM_base64_encodef( fin, fout );	/* encode and output */

	fclose(fin);	/* close the input file */
	fclose(fout);	/* close the output file */

	return 0;
}








/*------------------------------------------------------------------------
Procedure:     AM_attachment_replace ID:1
Purpose:       Replace the attachment in a mailbody
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_attachment_replace( char *mpack_name, char *old_att_name, char *new_att_name, char *att_path )
{
	FILE *in, *tmp, *newa;
	char *ofn, *nfn;
	char line[AM_1K_BUFFER_SIZE];
	char lline[AM_1K_BUFFER_SIZE];
	char lname[AM_1K_BUFFER_SIZE];
	char ts[AM_1K_BUFFER_SIZE];
	char tname[AM_1K_BUFFER_SIZE];
	char *p, *q;
	int base64hit = 0;
	int hit = 0;
	int quote_on = 0;
	int namesmatch = 0;
	int i = 0;
	char *index;
	char *nameindex = NULL;
	char *fnameindex = NULL;

	in = fopen( mpack_name, "r" ); /* open up our original mailpack */

	if (!in)	/* handle file opening error */
	{
		LOGGER_log("AM_attachment_replace: Cannot open mailpack %s for reading, \"%s\""\
				,mpack_name\
				,strerror(errno));
		return -1;
	}


	// First we must take the old file name, and translate it into
	// something we can use to compare against after it has been
	// brutalised and reduced (for safety)
	//
	ofn = strrchr(old_att_name,'/');  	/* get the attachment name out of the full path */
	if (!ofn) ofn = old_att_name;			/* if there was no '/', then reset the ofn ptr */
	else ofn++; 								/* skip the last '/' */
	strcpy(lname, ofn); 						/* create a lowercase version */
	PLD_strlower(lname);
	FNFILTER_filter( lname, strlen(lname) );/* clean up the filename */

	// Get the new attachment name out of a full path
	// This is simply so that we can insert this new name
	// into the headers without the pointless path prefix.
	//
	nfn = strrchr(new_att_name,'/');		/* extract new att name from path */
	if (!nfn) nfn = new_att_name;			/* if there was no '/' then reset the nfn ptr */
	else nfn++;


	if (glb.debug) LOGGER_log("AM_replace(): looking to replace %s with %s, having %s and %s", ofn, nfn, mpack_name, att_path );

	// Create the temporary mailpack name
	// This will become the new mail pack
	// assuming that everything goes okay in the end.
	//
	sprintf(tname,"%s.tmp", mpack_name);/* create the temporary mailpack name */
	tmp = fopen( tname, "w" );				/* try open the temp file */
	if (!tmp) /* handle file opening error */
	{
		LOGGER_log("AM_attachment_replace: Cannot open temporary file %s for writing, \"%s\""\
				,tname\
				,strerror(errno));
		return -1;
	}


	// Into the mailpack, first we must read in the headers, this must
	// be done in memory, so that should we encounter any of the required
	// header details, we dont have to try and back-track through a file
	// This gets rather awkward, as we dont know (often) if the file is
	// replacable before we get to its first occurance of the filename
	// header. This means we often have to backtrack ( a real PAIN ).
	//
	while (!feof(in))
	{
		hit = 0;
		base64hit = 0;
		quote_on = 0;
		namesmatch = 0;
		nameindex = NULL;
		fnameindex = NULL;
		AM_hbuffer_reset();// prepare the header buffers


		// Read in our file lines, searching for our attachment
		// which we are supposed to be replacing.
		//
		while (fgets(line, AM_1K_BUFFER_SIZE, in))
		{

			// Store this line into the header-buffer for later
			// use.
			//
			index = AM_hbuffer_add(line, tmp);

			strcpy(lline,line); /* copy and convert to lowercase the line we've read */
			PLD_strlower(lline);

			// If we detect "base64" on this line, then set the flag
			//
			if (strstr(lline,"base64"))
			{
				if (glb.debug) LOGGER_log("AM_replace(): BASE64 found - flag set.");
				base64hit=1;
			}

			// If we detect "filename=" on this line, then we need to
			// remember it, for later processing (if required)
			//
			if ((p=strstr(lline,"name="))) /* if the line contains name=, then check it */
			{
				if (!fnameindex) fnameindex = index;
				else nameindex = index;

				// Okay, so we've found the "name" line, now, lets
				// process the data on this line, and see if we have a
				// matching filename.
				//
				quote_on = 0;					/* no quotes found yet */
				p = p +strlen("name=");		/* move to just after the ='s */
				if (*p == '\"')				/* if there's a quote... */
				{
					quote_on++;					/* set quote to ON */
					p++;							/* move to next char */
				}

				q = strtok(p,";\"\n\r");	/* extract the filename */
				FNFILTER_filter( q, strlen(q) );

				// Compare both the length and contents of the
				// filenames.
				//
				if (\
						(strlen(q) == strlen(lname))\
						&& (strncmp(q,lname, strlen(lname))==0)\
					)
				{
					if (glb.debug) LOGGER_log("AM_replace(): Hit on filenames  %s and %s",q,lname);
					namesmatch = 1;
				}
				else
				{
					namesmatch = 0;
					base64hit = 0;
					fnameindex = 0;
					nameindex = 0;
				}
			}


			// Now, if we've found this is a base64 attachment and we
			// have also matched the names (matching done here)
			//
			if ( (base64hit == 1) && (namesmatch == 1) )
			{
				if (glb.debug) LOGGER_log("Base64 and NameMatch located, now fixing headers");
				// Okay, so we've gotten the match, and we need to replace
				// this filename details with those of the new file...
				// This is were we use the nameindex and fnameindex, for
				// from those two points, we alter the filenames
				//

				// First, lets do the "name" only one
				// strcpy the details so we can play with it safely
				//
				if (nameindex)
				{
					strcpy(line, nameindex);
					PLD_strlower(line);

					// 2. find where the "name" bit starts, and \0 it there
					//
					p = strstr(line,"name=") +strlen("name=");

					// 3. If the filename was quoted, remember this, and move one
					// 	char off to the right.
					if (*p == '"')
					{
						p++;
						quote_on=1;
					}
					else quote_on=0;
					*p = '\0';

					// 4. The actual filename is now ahead of us, until the next
					//		quote or \n\r.  We in effect, can ignore this now, and
					//		construct our new line.
					//
					sprintf(lline,"%s%s",line,nfn);
					if (quote_on) strcat(lline,"\"");
					strcat(lline,"\n");

					// 5. Copy this created string back, this requies a little
					//
					strncpy(nameindex,lline,1023);
				}

				//
				// The end of the filename swapout procedure.
				//
				// Now, do the same for the "filename" line....
				//
				if (fnameindex)
				{
					strcpy(line, fnameindex);
					PLD_strlower(line);
					p = strstr(line,"name=") +strlen("name=");
					if (*p == '"')
					{
						p++;
						quote_on=1;
					}
					else quote_on=0;
					*p = '\0';
					sprintf(lline,"%s%s",line,nfn);
					if (quote_on) strcat(lline,"\"");
					strcat(lline,"\n");
					strncpy(fnameindex,lline,1023);
				}
				// We are now happy with the way the headers are in the
				// memory, we can dump them to the output file now.
				//
				hit++;						/* flag that we had success */
			}

			if (hit && base64hit) break;


		} // WHILE fgets(line)

		// Write out the headers which we have been so faithfully
		// keeping in our holds
		//
		for (i = 0;
				i < AM_hbuffer_getmax();
				i++)
		{
			fprintf(tmp,"%s",AM_hbuffer_getline(i));
		}

		// Reset the buffers
		//
		AM_hbuffer_reset();


		/* If we had success with finding the right filename etc... */
		if ((hit)&&(base64hit))
		{
			hit = base64hit = 0;

			/* Step 2. Locate the start of the attachment */
			/* Step 2. Locate the start of the attachment */
			/* Step 2. Locate the start of the attachment */
			while (fgets(line, AM_1K_BUFFER_SIZE, in))
			{

				strcpy(lline,line);			/* copy and convert the next line */
				PLD_strlower(lline);

				if ((p=strstr(lline,"name=")))	/* if the line has name= in it... */
				{

					p = p +strlen("name="); /* get the position where the name starts */
					if (*p == '\"')			/* if quoted... */
					{
						quote_on=1;				/* set quote flag */
						p++;						/* step one char to the right */
					}
					else quote_on=0;

					q = strtok(p,";\"\n\r");	/* extract the filename */
					FNFILTER_filter( q, strlen(q) );

					if ( (strlen(q) == strlen(lname))\
							&& (strncmp(q,lname, strlen(lname))==0)
						)
					{
						p = q;
						q = line +(p -lline);
						*q = '\0';
						sprintf(ts, "%s%s", line, nfn);
						if (quote_on) strcat(ts,"\"");
						strcat(ts,"\n");
						strcpy(line,ts);
					}
				}


				fprintf(tmp,"%s",line);

				if (strlen(line) < 3) break;	/* if we've gotten to a blank line, exit */
			}


			/* Step 3. Insert attachment */
			/* Step 3. Insert attachment */
			/* Step 3. Insert attachment */

			newa = fopen( att_path, "r" );		/* Open the new attachment */

			if (!newa)										/* if we couldn't open it */
			{
				LOGGER_log("AM_attachment_replace: Cannot open file \"%s\" required to be inserted.(%s)"\
						,new_att_name, strerror(errno));

				/* as we couldn't open the new attachment, then we better keep the
				 * existing file so that we dont bugger up the mailpack, better to have
				 * the old attachment than non at all! */
				while (fgets(line, AM_1K_BUFFER_SIZE, in))
				{
					fprintf(tmp,"%s",line);
					if (strpbrk(line,"-")) break;
				} // While getting input file

			} // if New Attachment couldn't be opened

			else
			{
				if (glb.debug) LOGGER_log("AM_replace(): Starting to encode file...");
				AM_base64_encodef( newa, tmp );	/* Encode and output the new attachment */
				fclose(newa);								/* close the new attachment */
				if (glb.debug) LOGGER_log("AM_replace(): Encoding and insertion complete.");

				/* absorb the old attachment */
				while (fgets(line, AM_1K_BUFFER_SIZE, in))
				{
					/* we're assuming that in base64, only
					 * boundary lines have a '-' in them
					 * This is possibly a bad thing (IMHO)
					 * but short of tracking boundaries, I
					 * have no other way */
					if (strpbrk(line,"-"))
					{
						fprintf(tmp,"%s",line);
						break;
					} // If strpbrk

				} // While fgets

				if (glb.debug) LOGGER_log("AM_replace(): Attachment absorption complete.");

			} // ELSE

		} // While (major) fgets

	} // While !feof()

	clearerr(in);										/* clear out any errno's */
	fclose(tmp);										/* close our new mailpack */
	fclose(in);											/* close our old mailpack */
	rename( tname, mpack_name );					/* rename! */

	return 0;
}






/*------------------------------------------------------------------------
Procedure:     AM_add_section1 ID:1
Purpose:       Reads the main headers off an email and places the values it finds for
content-type, encoding and other such into the dd record
Input:         struct AM_disclaimer_details *dd : pointer to the record where all the details will go
FFGET_FILE *f : Open stream which we're reading the headers from
FILE *newf : Open stream which we're dumping the headers we read back into.
Output:
Errors:
------------------------------------------------------------------------*/
int AM_add_section1( struct AM_disclaimer_details *dd, FFGET_FILE *f, FILE *newf )
{
	char line[AM_1K_BUFFER_SIZE +1];
	char lline[AM_1K_BUFFER_SIZE +1];
	char *pos, *newpos, *tmppos;

	dd->isfile = 0;
	dd->ishtml = 0;

	if (FFGET_feof(f)) return -1;

	// Keep reading in lines until we have read a true-blank ( indicating end of headers )

	while ( FFGET_fgets(line, AM_1K_BUFFER_SIZE +1, f) )
	{
		fprintf(newf,"%s",line);
		if (glb.debug) LOGGER_log("%s:%d:AM_add_section1:IN:%s",FL,line);
		if (f->trueblank)
		{
			// If the line read in was a /TRUE-BLANK/, then we have
			// reached the end of our headers, so, it's time to
			// move on to the next section.
			if (glb.debug) LOGGER_log("%s:%d:AM_add_section1: True-BLANK line found, stopping header reading",FL);
			break;
		}

		strcpy(lline,line);
		PLD_strlower(lline);

		// Look for various header components in the lines we just read in
		//

		// Test for a filename

		if ( (strstr(lline,"name="))||(strstr(lline,"name =")) )
		{
			dd->isfile = 1;
		}

		// Test for content-type
		//
		// We need to determine the content-type so that we can test to see if
		// we've got a plain-text or a HTML body involved
		//

		if ((pos=strstr(lline, "content-type:")))
		{

			if (strstr(lline,"/signed"))
			{
				if (glb.alter_signed == 0)
				{
					if (glb.verbose) LOGGER_log("Message is signed, not altering.\n");
					return AM_RETURN_SIGNED_EMAIL;
				}
			}

			pos += strlen("content-type:");	// move our pointer along

			tmppos = pos;
			if ((pos=strstr(tmppos,"text/")))	// do we have a text type
			{
				// this is a text section

				dd->content_type = AM_contenttype_text_plain;

				// We honestly dont "care" about the complexities of the text included
				// other than if it's  HTML

				pos += strlen("text/");	// move pointer along

				if (strncmp(pos,"html",4)==0) dd->content_type = AM_contenttype_text_html;

			}
			else if ((pos=strstr(tmppos,"multipart/")))
			{
				dd->content_type = AM_contenttype_text_multipart;
			}
			else
			{
				dd->content_type = AM_contenttype_unknown;
			}

			if (glb.debug) LOGGER_log("%s:%d:AM_add_section1:Content line found",FL);

		} // Content-type line



		// Test for Content-Transfer-Encoding:
		//
		// CTE indicates what sort of format the data is encoded in in the body
		// As we're only interested in dealing with /text/ based content, it's
		// important that we find out what encoding we have.
		//
		// There is little point trying to add a disclaimer to something which
		// is BASE64 encoded!
		//

		if ((pos=strstr(lline, "content-transfer-encoding:")))
		{
			pos += strlen("content-transfer-encoding:");
			if (strstr(pos,"base64")) dd->content_encoding = AM_encoding_base64;
			else if (strstr(pos,"quote")) dd->content_encoding = AM_encoding_print;
			else if (strstr(pos,"7bit")) dd->content_encoding = AM_encoding_7bit;
			if (glb.debug) LOGGER_log("%s:%d:AM_add_section1:Content encoding found: %d", FL, dd->content_encoding);
		} // If line was a CTE




		// Test for Boundary specifier
		//
		// A very important line to detect for.  Without the boundary specifier, we
		// cannot correctly parse through a MIME encoded email.
		//

		if (dd->boundary_found == 0)
		{
			pos = strstr(lline, "boundary=");
			if (pos)
			{
				newpos = strstr(lline,"subject");
				if ((!newpos)||(newpos > pos))
				{
					if (glb.debug) LOGGER_log("%s:%d:AM_add_section1:Possible boundary (%s)",FL,pos);
					newpos = (pos -lline) +line +strlen("boundary=");	// move the pos pointer to just after the =

					pos = newpos;

					if ('\0' != *(pos +1))
					{
						if (*pos == '\"') pos++;		// If the boundary is quoted, then move along one char

						newpos = strtok(pos,"\n\r\"");	// Tokenise out the boundary string

						if (newpos)
						{
							strcpy(dd->boundary, newpos);		// copy it to our static string array

							dd->boundary_found = 1;				// Set the boundary_found flag

							if (glb.debug) LOGGER_log("%s:%d:AM_add_section1:boundary located",FL);
						}
					}

				} // If boundary is -genuine-

			} // If boundary located

		} // If we don't already have a boundary

	} // WHILE more lines to read or we have not found a true-BLANK

	if (glb.debug) LOGGER_log("%s:%d:AM_add_section1:Exit ( Header read section ).\n\t-- isfile=%d ishtml=%d boundaryfound=%d\n\n", FL, dd->isfile, dd->ishtml, dd->boundary_found );
	return 0;
}



/*------------------------------------------------------------------------
Procedure:     AM_disclaimer_load_text ID:1
Purpose:       Loads the text from the file given in fname, and mallocs the required amount
of memory to place the entire contents of the file into textptr.
Input:         char *fname : The file name from which to read the dislcaimer text
char **textptr : a derefernced pointer to the buffer variable which will hold the text.
Output:
Errors:
------------------------------------------------------------------------*/
int AM_disclaimer_load_text( char *fname, char **textptr )
{
	FILE *f;
	struct stat st;

	if (0 == stat(fname, &st))
	{
		*textptr = malloc(st.st_size +1);  // We have to add 1 so that we delimit the data with a \0

		if (textptr)
		{

			memset(*textptr, '\0', (st.st_size +1));

			f = fopen(fname,"r");
			if (!f)
			{
				LOGGER_log("%s:%d:ERROR - Cannot open %s for reading (%s)",__FILE__,__LINE__,fname,strerror(errno));
				textptr = NULL;
				return 0;
			}

			fread(*textptr, 1, st.st_size, f);
			if (glb.debug) LOGGER_log("Disclaimer Loaded:\n%s",*textptr);
			fclose(f);
		}
	}
	else
	{
		LOGGER_log("%s:%d:ERROR - Cannot stat '%s' (%s)",__FILE__,__LINE__,fname,strerror(errno));
	}

	return 0;
}




/*------------------------------------------------------------------------
Procedure:     AM_add_disclaimer_insert_html ID:1
Purpose:       Inserts a disclaimer with <PRE> tagging into a HTML block of text just prior to the
end of the HTML. Assumes that it has been placed at the start of the HTML block
Input:         AM_disclaimer_details *dd : Disclaimer tracking information
FFGET_FILE *f : Input file
FILE *newf : Output file
Output:
Errors:
------------------------------------------------------------------------*/
int AM_add_disclaimer_insert_html( 	struct AM_disclaimer_details *dd, FFGET_FILE *f, FILE *newf )
{
	char line[ AM_1K_BUFFER_SIZE +1];
	char lline[ AM_1K_BUFFER_SIZE +1];
	char *prebody, *tmpbody;

	while (FFGET_fgets(line, AM_1K_BUFFER_SIZE, f))
	{
		strcpy(lline,line);
		PLD_strlower(lline);

		// Look for either a BODY or HTML closing tag.

		prebody = strstr(lline,"</body>");
		if (!prebody) prebody = strstr(lline,"</html>");

		// If we found one of the tags, then insert our disclaimer aruond about
		// here.

		if (prebody != NULL)
		{
			if (glb.debug) LOGGER_log("Inserting html-body disclaimer...");

			dd->html_inserted = 1;

			// prepare to print out the original up to the </body> or </html> part

			tmpbody = line +(prebody -lline);
			*tmpbody = '\0';

			// save to file the first part of the line segment

			if ( dd->disclaimer_text_HTML == NULL )
			{
				fprintf(newf,"%s<BR><PRE>\n%s</pre><BR>",line, dd->disclaimer_text_plain);
			} else {
				fprintf(newf,"%s<BR>", dd->disclaimer_text_HTML );
			}

			fprintf(newf,"<%s",(tmpbody+1));

			break;
		}
		else
		{
			fprintf(newf,"%s",line);
		}

	} // While FFGET_fgets()

	return dd->html_inserted;
}




int AM_add_disclaimer_cleanup( FILE *mp, FILE *newf, char *mpacktmp, char *mpackname )
{

	// Close our files

	fclose(mp);
	fclose(newf);

	// rename the files

	rename(mpacktmp, mpackname);

	return 0;
}


/*------------------------------------------------------------------------
Procedure:     AM_add_disclaimer ID:1
Purpose:       Adds a disclaimer to [potentially] plain-text, mixed/multipart and HTML emails.
Input:         char *mpackname: Mailpack/email name we're modifying
char *disclaimer: either the text, or the filename of the text we want to insert
int istext: 0 = dislciamer is text, 1 = disclaimer is filename
Output:
Errors:
------------------------------------------------------------------------*/
int AM_add_disclaimer( char *mpackname, char *disclaimer, int istext )
{
	FILE *mp, *newf;
	FFGET_FILE f;
	char line[AM_1K_BUFFER_SIZE+1]="";
	char mpacktmp[AM_1K_BUFFER_SIZE+1]="";
	char mpackold[AM_1K_BUFFER_SIZE+1]="";
	struct AM_disclaimer_details dd;
	int result = 0;
	int last_boundary_written = -1;

	/* create our temp filename */
	snprintf(mpacktmp,AM_1K_BUFFER_SIZE, "%s.tmp",mpackname);
	snprintf(mpackold,AM_1K_BUFFER_SIZE, "%s.old",mpackname);

	/* Initialise our files */
	newf = fopen(mpacktmp,"w");
	mp = fopen(mpackname,"r");


	// Test file statuses... and hop out if we had troubles with either
	//
	if (!newf)
	{
		LOGGER_log("AM_add_disclaimer: Cannot open %s, %s",mpacktmp,strerror(errno));
		return -1;
	}

	if (!mp)
	{
		LOGGER_log("AM_add_disclaimer: Cannot open %s, %s",mpacktmp,strerror(errno));
		return -1;
	}


	// Set up the disclaimer text as required to be inserted into the file
	//

	if ((glb.disclaimer_plain != NULL )&&(glb.disclaimer_plain_type != AM_DISCLAIMER_TYPE_NONE))
	{
		if (glb.disclaimer_plain_type == AM_DISCLAIMER_TYPE_FILENAME)
		{
			AM_disclaimer_load_text( glb.disclaimer_plain, &(dd.disclaimer_text_plain) );
		}
		else
		{
			dd.disclaimer_text_plain = glb.disclaimer_plain;
		}
	} else {
		LOGGER_log("AM_add_disclaimer: Plain-text disclaimer has not been setup correctly\n");
	}


	if ((glb.disclaimer_HTML != NULL )&&(glb.disclaimer_HTML_type != AM_DISCLAIMER_TYPE_NONE))
	{
		if (glb.disclaimer_HTML_type == AM_DISCLAIMER_TYPE_FILENAME)
		{
			AM_disclaimer_load_text( glb.disclaimer_HTML, &(dd.disclaimer_text_HTML) );
		}
		else
		{
			dd.disclaimer_text_HTML = glb.disclaimer_HTML;
		}
	} else {
		LOGGER_log("AM_add_disclaimer: HTML-text disclaimer has not been setup correctly\n");
	}


	// Initialise our variables

	dd.content_type = AM_contenttype_none;
	dd.content_encoding = AM_encoding_none;
	dd.boundary_found = 0;
	memset(dd.boundary, 0, sizeof(dd.boundary));
	dd.ishtml = 0;
	dd.isfile = 0;
	dd.text_inserted = 0;
	dd.html_inserted = 0;


	if (glb.verbose) LOGGER_log("Attempting to add disclaimer");

	FFGET_setstream(&f, mp);

	// The process....
	//
	//	First, we read through the email until we at least come across the traditional
	//  true-blank seperator line between the headers and the body of the email.
	//  Once we have found that, we can then choose how we're going to detect our first
	//  disclaimer position based on what we discovered in the headers (ie, did we
	//  find a boundary specifier? are we using plain-text here, or does the email have
	//  HTML in it?
	//
	//  If we have just a plain text email, then we wait till we reach the end of the email
	//  and we just append the disclaimer
	//
	//  If we [most likely] dont have a plain email, then we need to wait until we find
	//  the boundary line, then, insert the disclaimer there.
	//
	//  Next, we need to check to see if the NEXT section is the HTML version of the first
	//  section (well, we dont /really/ compare them, but we check the headers to see if
	//  it has things like text/html and no filename
	//  If the section is HTML and non-filenamed, we once more wait until the boundary then
	//  we add in a HTML version of the disclaimer.
	//
	// Sounds easy ?
	//

	// Section 1 - read in the MAIN headers until a true-blank.
	//		This will give us some indication of what to expect in the email, be it a multipart
	//		a file, or a plain text email.
	//

	if (glb.debug) LOGGER_log("Reading main headers\n");

	result = AM_add_section1( &dd, &f, newf );

	if ((AM_RETURN_SIGNED_EMAIL == result)||(-1 == result))
	{
		fclose(mp);
		fclose(newf);
		if (glb.verbose) LOGGER_log("Email is signed with a crypto, we do not alter these emails\n");
		return AM_RETURN_SIGNED_EMAIL;
	}


	// Section 2 - Look for the end of the first text-body
	//
	// Now, if we have no boundary on this email, and it's of
	// text/* content-type
	//

	// If we have no boundary, and we have either text or no content-type
	// then we can just append the disclaimer to the /end/ of the email

	if ((0 == dd.boundary_found)&&((dd.content_type == AM_contenttype_text_plain)||(dd.content_type == AM_contenttype_text_html)))
	{
		if (glb.debug) LOGGER_log("Inserting disclaimer into a non-boundary email\n");

		// If we have a HTML email body, then insert the HTML based
		//		disclaimer, or, the plain-text one with <PRE>...</pre>
		//		tagging

		if (dd.content_type == AM_contenttype_text_html)
		{
			if (glb.debug) LOGGER_log("Conditions right for HTML disclaimer to be added\n");

			AM_add_disclaimer_insert_html(  &dd, &f, newf );
			if ((glb.verbose)&&( dd.html_inserted == 0 ))
			{
				LOGGER_log("WARNING: Could not insert HTML disclaimer into email\n");
			}

		} else {

			// If we have a plain-text email body, then just get to the end of the file
			//		and append the disclaimer to the end.

			while (FFGET_fgets(line, sizeof(line), &f))
			{
				fprintf(newf,"%s",line);
			}
			fprintf(newf,"%s",dd.disclaimer_text_plain);

			dd.text_inserted = 1;
		}


	}


	// BOUNDARY LIMITED EMAILS -----------------
	//
	// These emails are somewhat /harder/ to insert attachments into
	// namely because of the profound number of convoulted ways in which
	// email clients can forward/reply/mangle emails.
	//
	// We use a 'WHILE' loop here just so that we can cheat a bit when
	// trying to /exit/ out of the processing prematurely when we run
	// out of input (as happens at the end of the input file.
	//
	// By using a 'WHILE', we can just use 'break', rather than resorting
	// to other ugly methods :)


	while (1 == dd.boundary_found)
	{
		if (FFGET_feof(&f)) break;

		// If we've found a boundary and a text content section...
		//

		if (glb.debug) LOGGER_log("%s:%d:AM_add_disclaimer:Inserting into a MIME email\n",FL);


		// Step 1 - first read off any spurilous data prior to the first
		// 	MIME block, this typically includes things like
		//   "If you can read this, then your email client is not MIME
		//	  compliant"  (etc).

		if (glb.debug) LOGGER_log("\t%s:%d:AM_add_disclaimer:Reading spurilous data before first MIME segment\n",FL);

		while (FFGET_fgets(line, AM_1K_BUFFER_SIZE, &f))
		{
			fprintf(newf,"%s",line);
			if ( strstr(line,dd.boundary) ) break;
		}


		// IF for some reason we run into the end of the file, then we just quit out
		//	using our break.

		if (FFGET_feof(&f))
		{
			break;
		}


		// Okay, while we haven't yet managed to insert the disclaimer text
		//	or while we haven't inserted the HTML disclaimer text, we will
		//	keep going around in this loop

		while ((0 == dd.text_inserted)||(0 == dd.html_inserted))
		{

			// Step 2 - Now that we have the first block out of the way, we
			// can read to the bottom of the 2nd block, which is / should be
			// our text body


			// Read the headers and get the content-type information

			if (AM_add_section1( &dd, &f, newf ) < 0)
			{
				break;
				if (glb.verbose) LOGGER_log("\tRan out of input whilst processing headers\n");
			}

			// If this segment is a file, then rather than sifting through anything else
			//	just read until we get to the next boundary line and 'continue' on from
			//	the start of the WHILE loop again.

			if ( dd.isfile == 1 )
			{
				if (glb.debug) LOGGER_log("\t%s:%d:AM_add_disclaimer:Jumping this MIME section as it is a FILE\n",FL);

				while (FFGET_fgets(line, AM_1K_BUFFER_SIZE, &f))
				{

					//					if (FFGET_feof( &f ))
					//					{
					//						fprintf(stdout,"debug: Lastline = --'%s'--", line );
					//					}

					fprintf( newf, "%s", line );
					if ( strstr( line,dd.boundary ) ) break;
				}

				if (glb.debug) LOGGER_log("\t%s:%d:AM_add_disclaimer:File writing DONE.\n",FL);

				// IF for some reason we run into the end of the file, then we just quit out
				//	using our break.

				if (FFGET_feof(&f))
				{
					break;
				}

				continue;
			}


			// based on our findings from the header reading,
			//	If our MIME block is plain text...
			//
			// Also, take note that if the MIME block is for a file, we dont
			//	insert into it ( else our disclaimer will potentially become
			//	part of the file, really clever stuff!

			if ( (0 == dd.text_inserted) \
					&& (dd.content_type == AM_contenttype_text_plain) \
					&& (dd.isfile == 0) \
				)
			{
				if (glb.debug) LOGGER_log("\t%s:%d:AM_add_disclaimer:Conditions right to insert disclaimer\n",FL);

				// Read in the data (line by line) until we locate the
				// boundary line, then, we write out the disclaimer text
				// -before- we write the disclaimer line down.

				if (glb.debug) LOGGER_log("\t%s:%d:AM_add_disclaimer:Reading first segment looking for boundary line\n",FL);

				// Read and write data until we locate the boundary line.
				//	NOTE - we -deliberately- break before writing the boundary
				//	line because we want to insert the disclaimer /before/
				//	we write the boundary line.

				while (FFGET_fgets(line, AM_1K_BUFFER_SIZE, &f))
				{
					last_boundary_written = 0;
					if ( strstr( line,dd.boundary ) ) break;
					fprintf( newf, "%s", line );
					last_boundary_written = 1;
				}

				// If we got to the end of the file, then, well, ooops, I guess
				//	we'll have to break out.

				if ( FFGET_feof( &f ) )
				{
					break;
				}


				// Assuming that if the file is NOT encoded in base64, then we can attempt
				//	to insert the data. - This is a bit of a assumptious test, because we're
				//	going on the fact that our outer while-loop did a test to make sure that
				//	the current MIME section is NOT a file, therefore, we're saying
				//	"This segment is not a file and is not B64 encoded, so, we can now
				//	insert some disclaimer text into it"

				if (dd.content_encoding != AM_encoding_base64)
				{
					if (glb.debug) LOGGER_log("\t%s:%d:AM_add_disclaimer:Inserting TEXT disclaimer (%s)\n", FL, dd.disclaimer_text_plain);

					fprintf( newf, "%s\n", dd.disclaimer_text_plain );

					dd.text_inserted = 1; // Our disclaimer has been inserted, so set the flag

				} // if the content encoding is not B64

				if ( last_boundary_written == 0 )
				{
					fprintf( newf, "%s", line );
					last_boundary_written = 1;
				}

			} // If text not inserted and the segment is not a file


			// Add in the HTML disclaimer (if wanted)

			if ( ( glb.HTML_too ) \
					&& ( 0 == dd.html_inserted ) \
					&& ( dd.content_type == AM_contenttype_text_html ) \
					&& ( dd.isfile == 0 ) \
				)
			{
				if (glb.debug) LOGGER_log("\t%s:%d:AM_add_disclaimer:Conditions right for HTML disclaimer to be added",FL);

				result = AM_add_disclaimer_insert_html(  &dd, &f, newf );
				if ((glb.verbose)&&(0 == result))
				{
					LOGGER_log("WARNING: Could not insert HTML disclaimer into email");
				}


			} // If we also want to include a disclaimer in HTML body

		} // If conditions are right to insert our disclaimer into this email with boundaries




		// IF for some reason we run into the end of the file, then we just quit out
		//	using our break.

		if (FFGET_feof(&f))
		{

			if (glb.debug) LOGGER_log("\t%s:%d:AM_add_disclaimer:Main boundary outter-loop FEOF detector hit, breaking out",FL);
			break;
		}

		// Clean up any remaining lines in the email

		while (FFGET_fgets(line, AM_1K_BUFFER_SIZE, &f)){
			fprintf(newf,"%s",line);
			//			LOGGER_log("DEBUG: %s",line );
		}

	} // If a boundary was found in the main headers


	// Clean up any remaining lines in the email
	if ( ! FFGET_feof( &f ) )
	{
		while (FFGET_fgets(line, AM_1K_BUFFER_SIZE, &f)) fprintf(newf,"%s",line);
	}


	// if we didn't get the disclaimer inserted, and this email has no boundary
	//	and there's no file-headers, then we can insert the plain text disclaimer
	//	at the end.


	if ( (dd.text_inserted == 0) && (dd.html_inserted == 0) && ( dd.isfile == 0 ) )
	{
		if (glb.verbose) LOGGER_log("\tNo disclaimer added, despite going through entire file.");

		if ( dd.boundary_found == 0 )
		{
			fprintf(newf,"\n%s\n\n",dd.disclaimer_text_plain);
		}
	}

	if (glb.verbose) LOGGER_log("Done.\n");


	AM_add_disclaimer_cleanup( mp, newf, mpacktmp, mpackname);


	// Free the malloc'd text if required
	if ((glb.disclaimer_plain_type == AM_DISCLAIMER_TYPE_FILENAME)&&(dd.disclaimer_text_plain))
	{
		free(dd.disclaimer_text_plain);
	}

	if ((glb.disclaimer_HTML_type == AM_DISCLAIMER_TYPE_FILENAME )&&(dd.disclaimer_text_HTML))
	{
		free(dd.disclaimer_text_HTML);
	}

	return result;
}






/*------------------------------------------------------------------------
Procedure:     AM_filename_fix ID:1
Purpose:       Converts a MIME header
Input:         line: string with the filename
removed_prefix: filename prefix to use
removed_count: the number of attachments we've processed / ID
Output:
Errors:
------------------------------------------------------------------------*/
int AM_filename_fix( char *line, char *removed_prefix, int removed_count )
{
	char *pos;
	char lline[AM_1K_BUFFER_SIZE];

	/* process the current line, as it contains the filename, and replace the filename
	 * with a -removed- one */
	pos = strrchr(line,'='); /* locate where the filename starts */
	pos++; /* move one char along */
	*pos = '\0'; /* terminate the string */

	sprintf(lline,"%s\"%s%d\"\n",line,removed_prefix,removed_count); /* create a new string, in lline (scratch pad)*/
	strcpy(line,lline);

	return 0;
}






/*------------------------------------------------------------------------
Procedure:     AM_nullify_attachment ID:1
Purpose:       Removes any attachment within the mailpack which matches
(minimally) with the attachmentname parameter
Input:         mpackname - The mailpack name
attachmentname - The name of attachments which is to be removed
Output:        none
Errors:        If attachment wasn't found, or could not be removed, a non-zero return value.
------------------------------------------------------------------------*/
int AM_nullify_attachment( char *mpackname, char *attachmentname, char *removed_prefix )
{
	FILE *mp, *newf;
	char line[AM_1K_BUFFER_SIZE]="";
	char safe[AM_1K_BUFFER_SIZE]="";
	char lline[AM_1K_BUFFER_SIZE]="";
	char mpacktmp[AM_1K_BUFFER_SIZE]="";
	char mpackold[AM_1K_BUFFER_SIZE]="";
	char boundary[AM_1K_BUFFER_SIZE]="";

	int content_type = AM_contenttype_none;
	int content_encoding = AM_encoding_none;

	int result = 0;
	int boundary_found = 0;
	int filename_found = 0;
	int mimetype_found = 0;
	int removed_count = 0;
	char *pos, *newpos;

	if (glb.verbose) LOGGER_log("Removing attachment \"%s\"",attachmentname);

	/* create our temp filename */
	sprintf(mpacktmp,"%s.tmp",mpackname);
	sprintf(mpackold,"%s.old",mpackname);

	/* Initialise our files */
	if(strcmp(mpackname,"-") == 0) newf = fdopen(fileno(stdout),"w");
	else newf = fopen(mpacktmp,"w");

	if (!newf)
	{
		LOGGER_log("AM_nullifyattachment: Error: Cannot open temporary file for writing (%s) : %s",mpacktmp, strerror(errno));
		return -1;
	}

	if(strcmp(mpackname,"-") == 0) mp = fdopen(fileno(stdin),"r");
	else mp = fopen(mpackname,"r");
	if (!mp)
	{
		LOGGER_log("AM_nullifyattachment: Error: Cannot open temporary file for reading");
		return -1;
	}

	/* Initialise our variables */
	content_type = AM_contenttype_none;
	content_encoding = AM_encoding_none;

	PLD_strlower(attachmentname);

	/* Read every line from the original mailpack */
	while ( fgets(line,AM_1K_BUFFER_SIZE,mp) )
	{

		/* copy and convert into lower-case the input line */
		strcpy(lline,line);
		strcpy(safe,line);
		PLD_strlower(lline);

		/* check and see if we've found the filename... */

		if ( (pos=strstr(lline, "name="))!=NULL)
		{
			if ((mimetype_found) || (glb.nullify_all) ||
					(!strstr(attachmentname,"/") &&  (pos=strstr(lline,attachmentname)))  )
			{
				filename_found = 1;
				if (glb.debug) LOGGER_log("\nlocated attachment...");
			}
		}

		/* check and see if this is a content-type line */
		if ((pos=strstr(lline, "content-type:")))
		{
			if (strstr(attachmentname,"/") && strstr(lline,attachmentname)  )
			{
				mimetype_found = 1;
				if (glb.debug) LOGGER_log("\nlocated attachment...");
			}
			pos += strlen("content-type:");	/* move our pointer along */

			if ((pos=strstr(pos,"text/")))	/* do we have a text type */
			{
				pos += strlen("text/");	/* move pointer along */

				if (strncmp(pos, "plain",5)==0) content_type = AM_contenttype_text_plain;
				else
					if (strncmp(pos, "html",4)==0) content_type = AM_contenttype_text_html;
					else
						if (strncmp(pos, "print",5)==0) content_type = AM_contenttype_text_print;
						else content_type = AM_contenttype_text_unknown;

			}
			else {
				content_type = AM_contenttype_unknown;
			}
		}

		/* Our transfer encoding system indicates if we have a attachment type which is
		 * binary-encoded stuff or plain text */
		if ((pos=strstr(lline, "content-transfer-encoding:")))
		{
			pos += strlen("content-transfer-encoding:");
			if (strstr(pos,"base64")) content_encoding = AM_encoding_base64;
			else if (strstr(pos,"quote")) content_encoding = AM_encoding_print;
			else if (strstr(pos,"7bit")) content_encoding = AM_encoding_7bit;
		}


		if (boundary_found < 1)
		{ /* if we have not found our boundary yet */
			pos = strstr(lline, "boundary=");	/* check for our boundary */
			if (pos) { /* if we found one! */

				newpos = (pos -lline) +line +strlen("boundary=");	/* move the pos pointer to just after the = */
				pos = newpos;
				if (*pos == '\"') pos++;	/* if our 1st char is a quote, move one more */
				newpos = strtok(pos,"\n\r\"");	/* tokenise out the boundary */
				strcpy(boundary, newpos);	/* copy it to our static string array */
				boundary_found++;	/* increment the boundary_found flag */
				if (glb.debug) LOGGER_log("found boundary...");
			} /* if POS */
		} /* if boundary_found < 1 */

		else
		{ /* if boundary_found >= 1 */

			/* if we've found the filename, then we must process until we find the next boundary */
			if (filename_found == 1)
			{
				AM_filename_fix(line, removed_prefix, removed_count);
				fprintf(newf,"%s",line); /* output the new filename */
				if (glb.debug) LOGGER_log("renamed attachment...");

				/* first, locate our \n\r blank line */

				while ( (fgets(line, AM_1K_BUFFER_SIZE,mp)) && (strlen(line) >3) ){
					strcpy(lline,line);
					PLD_strlower(lline);

					if (strstr(lline,"name=")){
						AM_filename_fix(line, removed_prefix, removed_count);
					}

					fprintf(newf,"%s",line);
				}


				/* Now, until the boundary is found again, absorb all lines */
				while ( (fgets(line, AM_1K_BUFFER_SIZE,mp)) && (!strstr(line, boundary)) );

				/* leave an empty file */
				fprintf(newf,"\n\n");

				/* and, given that we will exit on the boundary line, we can just edit from this loop as we write it later */
				strcpy(safe, line);

				/* we are -done- */
				filename_found = 0;
				mimetype_found = 0;
				removed_count++;
				if (glb.debug) LOGGER_log("nullified.");
			} /* if filename_found == 1 */

		} /* if-ELSE */

		fprintf(newf,"%s",safe); /* save the original line to the output file */
		filename_found = 0;

	} /* WHILE more lines to read in from original */

	if (glb.verbose) LOGGER_log("\ndone.\n");
	/* Close our files */
	fclose(mp);
	fclose(newf);

	/* Rename files */
	rename(mpackname, mpackold);
	rename(mpacktmp, mpackname);

	return result;
}


/*------------------------------------------------------------------------
Procedure:     AM_insert_Xheader ID:1
Purpose:       Adds an X-header to the first set of headers in an email mailpack.
Input:         char *fname: mailpack name
char *xheader: header string
Output:
Errors:
------------------------------------------------------------------------*/
int AM_insert_Xheader( char *fname, char *xheader)
{

	/* Tempfile tmpfile() fix contributed by David DeMaagd - 29/01/2001 */

	char line[ AM_1K_BUFFER_SIZE +1];
	char tpn[ AM_1K_BUFFER_SIZE +1];
	int header_written = 0;
	int result = 0;
	struct stat st;
	FFGET_FILE f;
	FILE *fi;
	FILE *fo;

	// Create a temporary file name, but, so that we dont
	// overwrite an existing file, we must check the name using
	// stat() to see if we get a non-zero response.

	snprintf(tpn, AM_1K_BUFFER_SIZE, "%s",fname);

	do {
		if (strlen(tpn) < (sizeof(tpn) -2))
		{
			strcat(tpn,".");
		}
		else
		{
			LOGGER_log("ERROR:%s:%d: Temporary file name string buffer out of space!\n",__FILE__,__LINE__);
			return 1;
		}
		//		LOGGER_log("DEBUG:%s:%d: Testing filename %s\n",__FILE__,__LINE__,tpn);
	}
	while (0 == stat(tpn,&st));


	// Attempt to open up the temporary file in write mode.
	// If the open fails, then this whole operation fails.
	// Ensure there's lots of good logging here so that when something
	// does go wrong, at least people wont be left in the dark as to
	// what went on.
	//
	// Same applies for opening the source file
	//

	fo = fopen(tpn,"w");
	if (!fo)
	{
		LOGGER_log("Error: Cannot open temporary file for writing");
		return 1;
	}


	fi = fopen(fname,"r");
	if (!fi)
	{
		LOGGER_log("Error: Cannot open Mailpack file for reading, %s",fname);
		return 2;
	}

	//setup our FFGET stream

	FFGET_setstream(&f, fi);


	// Load and go through every line in the email file, testing it for "trueblank"
	// status.  Once a trueblank is found, we write our header, as this marks the
	// end of the headers (the trueblank).  Beyond that, we keep on reading and
	// writing because we obviously want to keep the entire email content.

	while (FFGET_fgets(line, AM_1K_BUFFER_SIZE ,&f))
	{

		// If we've found that the line we just read is a /true/ blank (as apposed to a carefully
		// crafted line length to catch out our fgets(), /and/ if we've not written in a header
		// then we will insert the header.

		if ((0 != f.trueblank)&&(0 == header_written))
		{
			fprintf(fo,"%s\n",xheader);
			header_written = 1;
		}

		// In all cases, we must write out the line we read in, even if it was the true-blank line
		// Otherwise we'll end up losing a the header seperation, and all of a sudden your emails
		// do not make any more sence to email programs.

		fprintf(fo,"%s",line);
	}

	// Close our input files


	fclose(fo);
	fclose(fi);

	// We now can rename the temporary file to the new file name.
	// NOTE - previously we removed then renamed, however, on checking
	// with the manpages, 'man 2 rename' it seems it's not required, so
	// rather than waste CPU cycles, we'll just go by the book, and use
	// only rename.  There are a couple of situations where rename can/will
	// fail, obviously such as if the original file is marked read-only, or
	// we do not have write permissions to it.


	if (rename(tpn,fname) == -1)
	{
		result = 1;
		LOGGER_log("ERROR while attempting to rename '%s' to '%s' (%s) ", tpn, fname, strerror(errno));
	}

	return result;
}




/*------------------------------------------------------------------------
Procedure:     AM_header_adjust ID:1
Purpose:       Adjusts an -existing- header 
Input:         
char *xheader: header string
Output:
Errors:
------------------------------------------------------------------------*/
int AM_alter_header( char *filename, char *header, char *change, int change_mode )
{

	char line[ AM_1K_BUFFER_SIZE +1];
	char tpn[ AM_1K_BUFFER_SIZE +1];
	int main_headers_complete = 0;
	int header_written = 0;
	int result = 0;
	struct stat st;
	FFGET_FILE f;
	FILE *fi;
	FILE *fo;

	// Create a temporary file name, but, so that we dont
	// overwrite an existing file, we must check the name using
	// stat() to see if we get a non-zero response.

	if (AM_DNORMAL) LOGGER_log("%s:%d:AM_alter_header:DEBUG: Start [ %s, %s, %s, %d ]", FL, filename, header, change, change_mode );

	snprintf(tpn, AM_1K_BUFFER_SIZE, "%s",filename);

	do {
		if (strlen(tpn) < (sizeof(tpn) -2))
		{
			strcat(tpn,".");
		}
		else
		{
			LOGGER_log("%s:%d:AM_header_adjust:ERROR: Temporary file name string buffer out of space!\n",FL);
			return 1;
		}
	}
	while (0 == stat(tpn,&st));

	if (AM_DNORMAL) LOGGER_log("%s:%d:AM_alter_header:DEBUG: Temporary filename = %s", FL, tpn );

	// Attempt to open up the temporary file in write mode.
	// If the open fails, then this whole operation fails.
	// Ensure there's lots of good logging here so that when something
	// does go wrong, at least people wont be left in the dark as to
	// what went on.
	//
	// Same applies for opening the source file
	//

	fo = fopen(tpn,"w");
	if (!fo)
	{
		LOGGER_log("%s:%d:AM_header_adjust:ERROR: Cannot open temporary file '%s' for writing",FL,tpn, strerror(errno));
		return 1;
	}


	fi = fopen(filename,"r");
	if (!fi)
	{
		LOGGER_log("%s:%d:AM_header_adjust:ERROR: Cannot open Mailpack file for reading, %s",FL,filename,strerror(errno));
		return 2;
	}

	//setup our FFGET stream

	FFGET_setstream(&f, fi);

	// Convert the header we're looking for into lower case so that 
	// we make it esaier to search for

	PLD_strlower( header );

	// Load and go through every line in the email file, testing it for "trueblank"
	// status.  Once a trueblank is found, we write our header, as this marks the
	// end of the headers (the trueblank).  Beyond that, we keep on reading and
	// writing because we obviously want to keep the entire email content.

	if (AM_DNORMAL) LOGGER_log("%s:%d:AM_header_adjust:DEBUG: Starting seek through file. header_written = %d", FL, header_written );
	while (FFGET_fgets(line, AM_1K_BUFFER_SIZE ,&f))
	{
		int line_written = 0;

		// If we find a true-blank line, then we note that the main-headers section is now
		//	complete - We do not process any headers after this, as the rest of the data
		//	is the email body/content [ which we do not wish to alter ].

		if ((main_headers_complete == 0)&&(f.trueblank == 1)) main_headers_complete = 1;


		// If we're still processing the main headers, and we've not written/altered our
		//	particular header which we wish to alter, then proceed to check this newly
		//	read line to see if it is the one we wish to alter,
		//
		//	Irrespective of if we alter this line or not, it will be written at the bottom
		//	of this while-loop.  NOTE - we don't write the line within the header-alteration
		//	"if" block, as this would only make things more complicated on exit ( ie, did we
		//	write the line, or are we okay... more logic tests, more chance of a bungle ).

		if ((main_headers_complete == 0)&&(header_written == 0))
		{
			char *p;
			char low_line[AM_1K_BUFFER_SIZE +1];

			if (AM_DNORMAL) LOGGER_log("%s:%d:AM_header_adjust: line=%s",FL, line);

			// Make a copy of the header line, and convert it to lowercase so that we
			//	can make a string comparison

			snprintf( low_line, AM_1K_BUFFER_SIZE, "%s", line );
			PLD_strlower( low_line );

			p = strstr(low_line, header);
			if (p != NULL)
			{
				char *first_colon_position;

				// Because it's quite possible to put a heaer line in which replicates itself, such as...
				//	Subject:  the Subject: is foo
				//	we need to first check that the header position we got from strstr is /BEFORE/ the first
				//	colon in the line.
				
				first_colon_position = strchr(low_line, ':');
				if ( p < first_colon_position )
				{
					char *header_start, *header_end;				
					char startc, finishc;

					// Convert the position in the low_line string over to that
					//	of the normal 'line' string.

					p = line +(p - low_line);

					if (AM_DNORMAL) LOGGER_log("%s:%d:AM_header_adjust:DEBUG: Located header line",FL);
					header_start = strchr(p,':');
					if ( header_start == NULL )
					{
						LOGGER_log("%s:%d:AM_header_adjust:WARNING: Could not locate terminating ':' character on header name (%s)",FL,p);
						continue;
					}
					header_start++;
					startc = *header_start;

					header_end = strpbrk( header_start, "\n;" );
					if ( header_end == NULL )
					{
						LOGGER_log("%s:%d:AM_header_adjust:WARNING: Could not locate end of header value (%s)",FL,header_start);
						continue;
					}
					finishc = *header_end;

					// We now have the start and end points of the header string, we now place in the additional 
					// data ( or entirely new data ) based on what the mode is

					switch ( change_mode ) {
						case AM_HEADER_ADJUST_MODE_PREFIX:
							*header_start = '\0';
							fprintf(fo,"%s %s%c%s", line, change, startc, header_start+1);
							header_written = 1;
							line_written = 1;
							if (AM_DNORMAL) LOGGER_log("%s:%d:AM_header_adjust:DEBUG: Prefix mode output written",FL);
							break;
						case AM_HEADER_ADJUST_MODE_SUFFIX:
							*header_end = '\0';
							fprintf(fo,"%s %s%c%s", line, change, finishc, header_end+1 );
							header_written = 1;
							line_written = 1;
							if (AM_DNORMAL) LOGGER_log("%s:%d:AM_header_adjust:DEBUG: Suffix mode output written",FL);
							break;
						case AM_HEADER_ADJUST_MODE_REPLACE:
							*header_start = '\0';
							fprintf(fo,"%s %s%s", line, change, header_end );
							header_written = 1;
							line_written = 1;
							if (AM_DNORMAL) LOGGER_log("%s:%d:AM_header_adjust:DEBUG: Replace mode output written",FL);
							break;

						default:
							LOGGER_log("%s:%d:AM_header_adjust:ERROR: Unknown header adjustment mode '%d'",FL, change_mode );
					} // Switch (change_mode)
				} // If p < first_colon_position

			} // If p != NULL, ie, we found a matching string for the header			
		} // If we're still in the headers and we've still not found a matching header 

		// In all cases, we must write out the line we read in, even if it was the true-blank line
		// Otherwise we'll end up losing a the header seperation, and all of a sudden your emails
		// do not make any more sence to email programs.

		if (line_written == 0) fprintf(fo,"%s",line);
	} // While we still have more lines in the file.

	// Close our input files


	fclose(fo);
	fclose(fi);

	// We now can rename the temporary file to the new file name.
	// NOTE - previously we removed then renamed, however, on checking
	// with the manpages, 'man 2 rename' it seems it's not required, so
	// rather than waste CPU cycles, we'll just go by the book, and use
	// only rename.  There are a couple of situations where rename can/will
	// fail, obviously such as if the original file is marked read-only, or
	// we do not have write permissions to it.


	if (rename(tpn,filename) == -1)
	{
		result = 1;
		LOGGER_log("%s:%d:AM_header_adjust:ERROR: while attempting to rename '%s' to '%s' (%s) ", FL, tpn, filename, strerror(errno));
	}

	return result;
}



//-----------------------------------END----


